/*
    WebMaker - KDE HTML Editor
    Copyright (C) 1998, 1999 Alexei Dets <dets@services.ru>
	
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.
			
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
					
    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <qdir.h>
#include <qfileinfo.h>
#include <qdialog.h>
#include <qlineedit.h>
#include <qlabel.h>
#include <qpushbutton.h>
#include <qlayout.h>
#include <kapp.h>
#include <kdir.h>
#include <kiconloader.h>
#include <kmsgbox.h>
#include "dirview.moc.h"

class KInputBox:public QDialog
{
public:
		KInputBox(const char *text, const char *value, QWidget *parent=0);
  inline const char* getText() {return editor->text();}
protected:
		QLineEdit *editor;
};

KInputBox::KInputBox(const char *text, const char *value, QWidget *parent):
		QDialog(parent, 0, TRUE)
{
		QVBoxLayout *vbox=new QVBoxLayout(this, 5);

		editor=new QLineEdit(this);
		editor->setText(value);
		editor->setMinimumSize(editor->sizeHint());
		QLabel *label=new QLabel(editor, text, this);
		label->setMinimumSize(label->sizeHint());
		vbox->addWidget(label);
		vbox->addWidget(editor);

		QHBoxLayout *hbox=new QHBoxLayout();
		vbox->addLayout(hbox);
	
		hbox->addStretch(1);
		QPushButton *ok=new QPushButton(i18n("&OK"), this);
		ok->setDefault(TRUE);
		ok->setMinimumSize(ok->sizeHint());
		hbox->addWidget(ok);
		QPushButton *cancel=new QPushButton(i18n("&Cancel"), this);
		cancel->setMinimumSize(cancel->sizeHint());
		hbox->addWidget(cancel);
		
		vbox->addStretch(1);
		
		vbox->activate();

		editor->setFocus();
		
		connect(ok, SIGNAL(clicked()), this, SLOT(accept()));
		connect(cancel, SIGNAL(clicked()), this, SLOT(reject()));
}
	
Directory::Directory( Directory * parent, const char * filename, bool showAllFiles )
		: QListViewItem( parent ), f(filename )
{
		p = parent;
		readable = TRUE;
		showAll = showAllFiles;
		initView();
}


Directory::Directory( QListView * parent, bool showAllFiles )
		: QListViewItem( parent ), f("/")
{
		p = 0;
		readable = TRUE;
		showAll = showAllFiles;
		initView();
}

void Directory::initView()
{
		htmlFilter = "*.*htm* *.php* *.asp *.cfm *.css *.jss *.js";
		textFilter = "*.txt";
		imageFilter = "*.gif *.jpg *.png *.GIF *.JPG *.PNG";
		backupFilter = "*~ ~* *.bak";
		KIconLoader *loader = kapp->getIconLoader();
		QPixmap pixmap(loader->loadMiniIcon("folder.xpm"));
		setPixmap (0, pixmap);
}

void Directory::setOpen( bool o )
{
		KIconLoader *loader = kapp->getIconLoader();
		QPixmap pixmap;
		QListViewItem *item;
		
		if ( o && !childCount() )
				{
						QString s( fullName() );
						QDir thisDir( s );
						if ( !thisDir.isReadable() )
								{
										readable = FALSE;
										return;
								}

						thisDir.setSorting(QDir::Name | QDir::DirsFirst);
						const QFileInfoList * files = thisDir.entryInfoList();
						if ( files )
								{
										QFileInfoListIterator it( *files );
										QFileInfo * f;
										f=it.toLast();
										while( (f=it.current()) != 0 )
												{
														--it;
														if ( f->fileName() == "." || f->fileName() == ".." )
																; // nothing
														else if ( f->isDir() )
																{
																		Directory *dir=new Directory( this, f->fileName(), showAll );
																		dir->setFilter(HTML, htmlFilter);
																		dir->setFilter(TEXT, textFilter);
																		dir->setFilter(IMAGE, imageFilter);
																		dir->setFilter(BACKUP, backupFilter);
																}
														else if ( f->isFile() )
																{
																		if (KDir::match(backupFilter, f->fileName()))
																				pixmap = loader->loadMiniIcon("trash.xpm");
																		else if (KDir::match(htmlFilter, f->fileName()))
																				pixmap = loader->loadMiniIcon("html.xpm");
																		else if (KDir::match(textFilter, f->fileName()))
																				pixmap = loader->loadMiniIcon("txt.xpm");
																		else if (KDir::match(imageFilter, f->fileName()))
																				pixmap = loader->loadMiniIcon("image.xpm");
																		else if (showAll)
																				pixmap = loader->loadMiniIcon("unknown.xpm");
																		else continue;
																		item = new QListViewItem( this, (const char *)f->fileName(), "File", 0);
																		item -> setPixmap(0, pixmap);
																}
														else
																{
																		item = new QListViewItem( this, (const char *)f->fileName(), "Special", 0 );
																		pixmap = loader->loadMiniIcon("unknown.xpm");
																		item -> setPixmap(0, pixmap);
																}
												}
								}
				}
		else
				{
						item = firstChild();
						while(item)
								{
										delete item;
										item = firstChild();
								}
				};
		QListViewItem::setOpen( o );
}


void Directory::setup()
{
		KDir thisDir(fullName());
		if (!showAll)
				{
						QString filter(htmlFilter+" "+textFilter+" "+imageFilter+" "+backupFilter);
						thisDir.setNameFilter(filter);
				};
		if (thisDir.count()>1)
				setExpandable(TRUE);
		else
				setExpandable(FALSE);
		int textHeight=listView()->fontMetrics().height() + listView()->itemMargin()*2;
		int iconHeight=16 + listView()->itemMargin()*2;
		setHeight(textHeight > iconHeight ? textHeight : iconHeight);
		widthChanged();
}

void Directory::setFilter(FilterTypes filterType, const QString &filterString)
{
		if (!filterString.isEmpty())
				{
						switch(filterType)
								{
								case HTML: htmlFilter=filterString;break;
								case TEXT: textFilter=filterString;break;
								case IMAGE: imageFilter=filterString;break;
								case BACKUP: backupFilter=filterString;break;
								}
				}
}

QString Directory::fullName()
{
		QString s;
		if ( p ) {
				s = p->fullName();
				s.append( f.name() );
				s.append( "/" );
		} else {
				s = "/";
		}
		return s;
}


const char * Directory::text( int column ) const
{
		if ( column == 0 )
				return f.name();
		else if ( readable )
				{
						return "Directory";
				}
		else
				return "Unreadable Directory";

}

DirView::DirView(bool showAllFiles, QWidget *parent, const char *name):QListView(parent, name)
{
		addColumn(i18n("Directory Tree"));
		//addColumn("Type");
		setStyle(WindowsStyle);
		setPalette(palette());
		setSorting(-1);
		showAll=showAllFiles;
		
		Directory *files=new Directory(this, showAllFiles);
		files->setOpen(TRUE);

		RBMenu = new QPopupMenu();
		openID=RBMenu->insertItem(i18n("&Open"), this, SLOT(openFile()));
		renameID=RBMenu->insertItem(i18n("Re&name"), this, SLOT(renameFile()));
		removeID=RBMenu->insertItem(i18n("&Remove"), this, SLOT(removeFile()));
		RBMenu->insertSeparator();
		mkdirID=RBMenu->insertItem(i18n("&New Directory"), this, SLOT(makeDir()));
		
		connect(this, SIGNAL(rightButtonPressed(QListViewItem*, const QPoint&, int)),
										this, SLOT(fileMenu(QListViewItem*, const QPoint&, int)));
}

void DirView::setStyle(GUIStyle)
{
		QListView::setStyle(WindowsStyle);
}

void DirView::setPalette(const QPalette &p)
{
		QListView::setPalette(p);
		if (QApplication::style()==WindowsStyle)
				QApplication::setWinStyleHighlightColor(kapp->selectColor);
		else
				QApplication::setWinStyleHighlightColor(kapp->textColor);
}

void DirView::setFilter(FilterTypes filterType, const QString &filterString, bool updateView)
{
		if (updateView)
				{
						delete firstChild();
						Directory *files=new Directory(this, showAll);
						files->setFilter(HTML, htmlFilter);
						files->setFilter(TEXT, textFilter);
						files->setFilter(IMAGE, imageFilter);
						files->setFilter(BACKUP, backupFilter);
						files->setOpen(TRUE);
				}
		else
				switch(filterType)
						{
						case HTML: htmlFilter=filterString;break;
						case TEXT: textFilter=filterString;break;
						case IMAGE: imageFilter=filterString;break;
						case BACKUP: backupFilter=filterString;break;
						}
}

void DirView::showAllFiles(bool enable)
{
		if (enable==showAll) return;
		showAll=!showAll;
		delete firstChild();
		Directory *files=new Directory(this, showAll);
		files->setOpen(TRUE);
}

QString DirView::getFileName(const QListViewItem *item)
{
		QString fileName;
		if (item->depth())
				for (int i=item->depth(); i>0; i--)
						{
								fileName.prepend(item->text(0));
								fileName.prepend("/");
								item=item->parent();
						}
		else
				fileName="/";
		return fileName;
}

void DirView::fileMenu(QListViewItem *item, const QPoint &point, int)
{
		setSelected(item, TRUE);
		QString fileName(getFileName(item));
		QFileInfo fileInfo(fileName);
		if (fileInfo.isDir())
				RBMenu->setItemEnabled(openID, FALSE);
		else
				RBMenu->setItemEnabled(openID, TRUE);
		RBMenu->popup(point);
}

void DirView::openFile()
{
		emit returnPressed(currentItem());
}

void DirView::renameFile()
{
		QString fileName(getFileName(currentItem()));
		KInputBox *dialog=new KInputBox(i18n("New Name:"), fileName, this);
		if (dialog->exec())
				{
						QDir dir;
						if (dir.rename(fileName, dialog->getText()))
								{
										QListViewItem *parent=currentItem()->parent();
										parent->setOpen(FALSE);
										parent->setOpen(TRUE);
								}
						else
								{
										KMsgBox::message(this, i18n("Error"), i18n("Error occured while renaming file!"),
																											KMsgBox::EXCLAMATION);
								}
				}
		delete dialog;													
}

void DirView::removeFile()
{
		QString fileName(getFileName(currentItem()));
		QFileInfo file(fileName);
		QString s;
		if (file.isDir())
				s.sprintf(i18n("Remove directory \"%s\"? Are you sure?"), file.fileName().data());
		else
				s.sprintf(i18n("Remove file \"%s\"? Are you sure?"), file.fileName().data());
		int result=KMsgBox::yesNo(this, i18n("Warning"), s);
		if (result==1)
				{
						QDir dir;
						if (file.isDir())
								result=dir.rmdir(fileName);
						else
								result=dir.remove(fileName);
						if (result)
								{
										QListViewItem *parent=currentItem()->parent();
										parent->setOpen(FALSE);
										parent->setOpen(TRUE);
								}
						else
								{
										KMsgBox::message(this, i18n("Error"), i18n("Error occured while removing file!"),
																											KMsgBox::EXCLAMATION);
								}
				}
}

void DirView::makeDir()
{
		QListViewItem *item=currentItem();
		QString fileName(getFileName(item));
		QFileInfo file(fileName);
		if (file.isFile() || !file.exists()) item=item->parent();
		fileName=getFileName(item)+"/";
		KInputBox *dialog=new KInputBox(i18n("New Directory Name:"), fileName, this);
		if (dialog->exec())
				{
						QDir dir;
						if (dir.mkdir(dialog->getText()))
								{
										item->setOpen(FALSE);
										item->setOpen(TRUE);
								}
						else
								{
										KMsgBox::message(this, i18n("Error"), i18n("Error occured while creating new directory!"),
																											KMsgBox::EXCLAMATION);
								}
				};
		delete dialog;
}
