/* $Id: kiojob.h,v 1.1.1.1 1998/05/06 19:13:31 matt Exp $
   $Log: kiojob.h,v $
   Revision 1.1.1.1  1998/05/06 19:13:31  matt
   Imported sources

   Revision 1.1  1998/03/02 22:00:19  matt
   Added resuming


   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

   */

#ifndef kiojob_h
#define kiojob_h

class KIOJob;
class KIOServer;

#include <qlist.h>
#include <qdialog.h>
#include <qstring.h>
#include <qlabel.h>
#include <qwidget.h>
#include <qdict.h>

#include <kurl.h>

#include "kstrlist.h"
#include "kioserver_ipc.h"

class KIOJob : public QObject
{
    Q_OBJECT
public:
    enum Jobs { JOB_COPY, JOB_RESUME };

    /**
     * Constructs a new job.
     * '_id' can be used to identify the job. Every signal transmitts this id,
     * so the client always knows which job sent him the signal.
     */
    KIOJob( int _id = 0 );
    ~KIOJob();
    
    void doIt( KIOSlaveIPC * _slave );

    void copy( QStrList & _src_url_list, const char *_dest_dir_url );    
    void copy( const char *_src_url, const char *_dest_url );

    void resume( QStrList & _src_url_list, const char *_dest_dir_url );    
    void resume( const char *_src_url, const char *_dest_url );

    /**
     * Deletes all running jobs and their slaves.
     * Do this only if you want to quit the application.
     */
    static void deleteAllJobs();

    /**
     * Allows the job to overwrite existing files during move and
     * copy commands.
     *
     * @see #overwriteExistingFiles
     */
    void KIOJob::setOverWriteExistingFiles( bool _o );

    /**
     * If this flag is set to TRUE, the instance deletes itself after
     * its job is finished. So you can start a KIOJob and dont have to care
     * about it any more. This flag is turned ON by default.
     */
    void setAutoDelete( bool _b ) { bAutoDelete = _b; }
    /**
     * @see #isAutoDelete
     */
    bool isAutoDelete() { return bAutoDelete; }

    /**
     * Call this function only if you received the signal @ref #errorFilter.
     * You can avoid further processing of this error with the help of this function.
     * 
     * @see #processError
     */
    void doNotProcessError() { bProcessError = false; }

    /**
     * If an error arrives @ref #fatalError is called and the error
     * is first emitted via the @ref errorFilter
     * signal. If as a result of this someone calls @ref #processError, then the
     * @ref #bProcessError flag is cleared to prevent calling this function
     * twice. If @ref #processError has not been called, then @ref #fatalError
     * calls it. One can use @ref #doNotProcessError to prevent @ref #fatalError
     * from doing so.
     */
    void processError( int _kioerror, const char* _url, int _errno );

    void msgErr( bool, QString );

    /*
     * hack to get static classes up and running even with C++-Compilers/
     * Systems where a constructor of a class element declared static
     * would never get called (Aix, Alpha,...). In the next versions these
     * elements should disappear.
     */
    static void InitStatic() {
    	passwordDict = new QDict<QString>;
    	jobList = new QList<KIOJob>;
    }

public slots:
    void slaveIsReady();
    void slaveStarted(); 
    void slaveProgress( int _percent, int _size );
    void cancel( bool try_later = false );
    void stop();
    void start( int _pid );
    void slotSlaveClosed( KIOSlaveIPC* );
    
    void slotData( IPCMemory _mem );
    /**
     * This slot is called if the kioslave produced some kind
     * of error. It emits the @ref #errorFilter signal and calls
     * @ref #processError afterwards, except a call to
     * @ref #doNotProcessError told us not to do so or if @ref processError
     * has already been called.
     */ 
    void fatalError( int _kioerror, const char* _url, int _errno );
    
    void slotRedirection( const char *_url );
    void slotInfo( const char *_text );
    
signals:
    void data( const char *_data, int _len );
    void redirection( const char *_url );
    void info( const char *_text );
    
    /**
     * Emitted if the job has finished.
     */
    void finished( int _id );
    void trylater( int _id );

    void jobstarted( int _id );

    void error( bool confirm, const char *_txt, int _id );
    
    /**
     * This signal is emitted if an error arrives and before it is processed.
     * You may call @ref #processError if you receive this signal. Use
     * @ref #doNotProcessError if you want to process the error on your own.
     *
     * @param _text is the debug text generated by kioslave. Dont show this string
     *              to the user. It is just for debugging.
     */
    void errorFilter( int _kioerror, const char *_text, int _errno );
  
    /**
     * Tells about the progress of this job in percent.
     *
     * @param _percent is between 0 and 100.
     */
    void progress( int _percent, int _size, int _id );
    
    /**
     * Tells about the amount of bytes read.
     * Not implemented yet.
     */
    void bytes( int _bytes );    

protected:
    /**
     * Looks wether the '_url' specifies a user but not a password. If so the
     * function will ask the user for a password if we dont know it already.
     * The a complete URL us assembled and returned.
     *
     * @return a complete URL with password if needed.
     *
     * @see #passwordDict
     */
    QString KIOJob::completeURL( const char *_url );
    
    void done( bool try_later = false );
    /**
     * This one makes recursive copy possible.
     * This function assumes that cmSrcURLList and cmDestURLList are filled
     * correctly. It searches in cmSrcURLList for directories. If it finds one,
     * it recursively traverses these and adds those files to the 2 lists.
     * After that is tells the server that this job is waiting for a slave.
     */
    void copy();

    void resume();
    
    KIOSlaveIPC *slave;

    /**
     * Used for COPY command.
     */
    KStrList cmSrcURLList;
    KStrList cmDestURLList;

    KStrList tmpSrcURLList;
    KStrList tmpDestURLList;
    
    
    /**
     * Used by DELETE/MOVE/COPY to store total count of files to process.
     */
    int cmCount;
        
    /**
     * Tells wether we already connected slots/signals to the slave.
     */
    bool started;

    KIOServer *server;

    /**
     * The action that is currently taking place.
     */
    Jobs action;

    /**
     * Flag indicated wether the slave had been told to clean up.
     * When the ob does not need the slave, the slave is told to clean up and
     * this flag is set. If the slave tells that he did the clean up, the dialog
     * window ( if any ) well be deleted and the job is deleted. But mention
     * that the cleanup may raise errors.
     */
    bool cleanedUp;

    /**
     * The last error.
     * If 'fatalError' is called, this variable holds the error value. This is
     * used later by 'msgResult'.
     */
    int kioError;

    /**
     * A unique ID.
     */
    int id;

    /**
     * List of all running jobs.
     */
    static QList<KIOJob> *jobList;

    /**
     * This flag is set to TRUE if existing files should be overwritten by copy/move jobs.
     */
    bool overwriteExistingFiles;
    
    /**
     * The keys for this dict look like "user@host".
     *
     * @see #completeURL
     */
    static QDict<QString> *passwordDict;

    /**
     * @see #setAutoDelete
     */
    bool bAutoDelete;

    /**
     * @see #processError
     * @see #fatalError
     */
    bool bProcessError;
};

#include "kioserver.h"

#endif
