/*
  a basic player
  Copyright (C) 1998  Martin Vogt

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation.

  For more information look at the file COPYRIGHT in this package

 */


#include <kmpgPlayer.h>



KmpgPlayer::KmpgPlayer(Playlist* playlist,QWidget *parent, const char *name) :
  QWidget(parent,name) { 

  this->playlist=playlist;
  filename=new Buffer(40);
  connect(playlist,SIGNAL(playSongEvent(int)),this,SLOT(playSongEvent(int)));
  setBackgroundColor (QColor(0,0,0));

  buildGraph();
  buildGUI();
}


KmpgPlayer::~KmpgPlayer() {
  // Before you delete a player remove the listener
  genericPlayer->removeListener(multicastDevice);
  delete genericPlayer;

  delete multicastDevice;
  delete spectrumAnalyserDevice;
  delete volumeDevice;
  delete streamGeneralInfoDevice;
  delete streamBufferInfoDevice;
  delete streamJumpInfoDevice;
  delete filename;
} 


QSize KmpgPlayer::sizeHint () {
  return QSize(275,120);
}


void KmpgPlayer::processEvent(char eventId) {
  switch(eventId) {
  case _GS_SIGNAL_PLAYING_READY: {
    playNext();
    break;
  }
  case _GS_SIGNAL_GENERATOR_CRASHED: {
    decoderCrash();
    break;
  }
  }
}
 

void KmpgPlayer::addListener(NodeDevice* listener) {
  multicastDevice->addListener(listener);
}


void KmpgPlayer::removeListener(NodeDevice* listener) {
  multicastDevice->removeListener(listener);
}


GenericPlayer* KmpgPlayer::getGenericPlayer() {
  return genericPlayer;
}


void KmpgPlayer::playSongEvent(int index) {
  Song* song=playlist->at(index);
  if (song != NULL) {
    KURL* url=song->getSong();
    filename->setData((char*)url->path());
    genericPlayer->open(filename->getData());
  }
}


void KmpgPlayer::jump(int second) {
  genericPlayer->jump(second);
}
 

void KmpgPlayer::play() {
  if (filename->len() == 0) {
    PlaylistOperation::playCurrent(playlist);
  } else {
    genericPlayer->play();
  }
}


void KmpgPlayer::pause() {
  genericPlayer->pause();
}

  
void KmpgPlayer::stop() {
  genericPlayer->close();
  if (GarbageCollector::canMakeCollection()) {
    GarbageCollector::makeGarbageCollection();
  }
}

void KmpgPlayer::playNext() {
 PlaylistOperation::playNext(playlist);
}


void KmpgPlayer::playPrev() {
  PlaylistOperation::playPrev(playlist);
}


void KmpgPlayer::decoderCrash() {
  Buffer* msg=new Buffer(50);
  msg->setData("Your decoder takes a trip to nirvana!\n\n");
  msg->append( "song: ");
  msg->append( filename->getData());
  msg->append("\n\ncrashprotection(tm)+autorestart(tm) v1.0");

  KMsgBox::message(this,"nirvana warning",msg->getData());
  delete msg;
  
  genericPlayer->removeListener(multicastDevice);
  delete genericPlayer;
  genericPlayer = Amplifier::createPlayer(_MP3_DECODER);
  genericPlayer->addListener(multicastDevice);
  connect(genericPlayer->getEventQueue(),SIGNAL(processEvent(char)),
	  this,SLOT(processEvent(char)));
  
  playNext();
  cout << "enter qt event loop"<<endl;

}


/**
   This constructs the internal device graph.
*/
void KmpgPlayer::buildGraph() {

  // First create the devices.
  genericPlayer = Amplifier::createPlayer(_MP3_DECODER);
  multicastDevice = new MulticastDevice();
  spectrumAnalyserDevice = new SpectrumAnalyserDevice();
  volumeDevice = new VolumeDevice();
  streamBufferInfoDevice = new StreamInfoDevice();
  streamGeneralInfoDevice = new StreamInfoDevice();
  streamJumpInfoDevice = new StreamInfoDevice();

  // Now we build the connection between them.
  multicastDevice->addListener(spectrumAnalyserDevice);
  multicastDevice->addListener(volumeDevice);
  multicastDevice->addListener(streamGeneralInfoDevice);
  multicastDevice->addListener(streamBufferInfoDevice);
  multicastDevice->addListener(streamJumpInfoDevice);
  genericPlayer->addListener(multicastDevice);

  // this device delivers message with "low" traffic:
  streamGeneralInfoDevice->setEventMask(_STREAMINFO_STATUS_CHANGE |
					_STREAMINFO_MUSIC_CHANGE  |
					_STREAMINFO_AUDIO_CHANGE );

  // these devices are for the high traffic message, use with care
  streamBufferInfoDevice->setEventMask(_STREAMINFO_BUFFERFILL_CHANGE);

  // the jumper needs its own device with its special eventmask
  streamJumpInfoDevice->setEventMask(_STREAMINFO_TIME_CHANGE   |
				     _STREAMINFO_STATUS_CHANGE |
				     _STREAMINFO_MUSIC_CHANGE ); 

  connect(genericPlayer->getEventQueue(),SIGNAL(processEvent(char)),
	  this,SLOT(processEvent(char)));
}

 


void KmpgPlayer::buildGUI() {
  QBoxLayout* topLayout = new QVBoxLayout( this, 1 );

  QBoxLayout* b = new QHBoxLayout();
  topLayout->addLayout(b);
  
  guiSpectrumAnalyser = new GuiSpectrumAnalyser(this);
  guiSpectrumAnalyser->attachTo(spectrumAnalyserDevice);
  b->addWidget(guiSpectrumAnalyser);

  
  statusWidget = new StatusWidget(this);
  statusWidget->attachTo(streamGeneralInfoDevice);
  b->addWidget(statusWidget);


  timerWidget = new TimerWidget(this);
  timerWidget->attachTo(streamJumpInfoDevice); // this device delivers "time"
  b->addWidget(timerWidget);


  QBoxLayout* b3 = new QVBoxLayout();
  b->addLayout(b3);

  
  volumeWidget = new VolumeWidget(this);
  volumeWidget->attachTo(volumeDevice);
  volumeWidget->buildGui();
  b3->addWidget(volumeWidget);
  

  guiBalanceDevice = new GuiBalanceDevice(this);
  guiBalanceDevice->buildGui();
  guiBalanceDevice->attachTo(volumeDevice);
  b3->addWidget(guiBalanceDevice);

  

  QBoxLayout* b1 = new QVBoxLayout();
  topLayout->addLayout(b1);


  infoWidget = new InfoWidget(this);
  infoWidget->attachTo(streamGeneralInfoDevice);
  b1->addWidget(infoWidget);


  guiJumpDevice = new GuiJumpDevice(this);
  guiJumpDevice->buildGui();
  guiJumpDevice->attachTo(streamJumpInfoDevice);
  b1->addWidget(guiJumpDevice);
  connect(guiJumpDevice,SIGNAL(jump(int)),this,SLOT(jump(int)));

  QBoxLayout* b2 = new QHBoxLayout();
  topLayout->addLayout(b2);

  controlWidget = new ControlWidget(this); 
  controlWidget->buildGui();
  b2->addWidget(controlWidget);
  connect(controlWidget,SIGNAL(playCurrentEvent()),this,SLOT(play()));
  connect(controlWidget,SIGNAL(playNextEvent()),this,SLOT(playNext()));
  connect(controlWidget,SIGNAL(playPrevEvent()),this,SLOT(playPrev()));
  connect(controlWidget,SIGNAL(playStopEvent()),this,SLOT(stop()));
  connect(controlWidget,SIGNAL(playPauseEvent()),this,SLOT(pause()));
  



  QBoxLayout* b4 = new QVBoxLayout();
  QBoxLayout* b5 = new QHBoxLayout();
  QBoxLayout* b6 = new QHBoxLayout();
  
  b2->addLayout(b4);
  b4->addLayout(b5);
  b4->addLayout(b6);
  

  guiBufferInfoDevice = new GuiBufferInfoDevice(this);
  guiBufferInfoDevice->attachTo(streamBufferInfoDevice);
  b5->addWidget(guiBufferInfoDevice);

  guiShuffle = new GuiShuffle("Rnd:",this);
  b5->addWidget(guiShuffle);
  guiShuffle->setChecked(playlist->getShuffle());
  connect(guiShuffle,SIGNAL(setShuffleEvent(int)),
	  playlist,SLOT(setShuffle(int)));
  connect(playlist,SIGNAL(setShuffleEvent(int)),
	  guiShuffle,SLOT(setShuffle(int)));
  
  guiRepeat = new GuiRepeat("Cnt:",this);
  b5->addWidget(guiRepeat);
  guiRepeat->setChecked(playlist->getRepeat());
  connect(guiRepeat,SIGNAL(setRepeatEvent(int)),
	  playlist,SLOT(setRepeat(int)));
  connect(playlist,SIGNAL(setRepeatEvent(int)),
	  guiRepeat,SLOT(setRepeat(int)));


  guiStereoDevice = new GuiStereoDevice(this);
  guiStereoDevice->attachTo(streamGeneralInfoDevice);
  b6->addWidget(guiStereoDevice);


  guiFreqDevice = new GuiFreqDevice(this);
  guiFreqDevice->attachTo(streamGeneralInfoDevice);
  b6->addWidget(guiFreqDevice);


  guiBPSDevice = new GuiBPSDevice(this);
  guiBPSDevice->attachTo(streamGeneralInfoDevice);
  b6->addWidget(guiBPSDevice);
}

 


