/*
  stores deviceConfig data in a fifo. (it makes a deep copy  of the data)
  Copyright (C) 1998  Martin Vogt

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation.

  For more information look at the file COPYRIGHT in this package

 */



#include <deviceConfig/deviceConfigFifo.h>



DeviceConfigFifo::DeviceConfigFifo(NodeDevice* id) {
  int i;

  freePos=0;
  lWaitForStoppedSource=false;
  lWaitForPausedSource=false;
  lDequeued=false;
  this->id=id;
  status=new StatusInfo();

  for(i=0;i<_MAX_FIFO_ENTRIES;i++) {
    configQueue[i]=new DeviceConfig();
  }
}


DeviceConfigFifo::~DeviceConfigFifo() {
  int i;
  
  clear();
  for(i=0;i<_MAX_FIFO_ENTRIES;i++) {
    delete configQueue[i];
  }
  delete status;
}


  

void DeviceConfigFifo::enqueue(DeviceConfig* config) {
  DeviceConfig* freePosConfig;
  MemChunk* memChunk;
  AudioBuffer* audioBuffer;

  if (freePos == _MAX_FIFO_ENTRIES) {
    cout << "fifo overflow! ignore enqueue"<<endl;
    return;
  }
  freePosConfig=configQueue[freePos];
  config->copyTo(freePosConfig);
  audioBuffer=config->getAudioBuffer();
  memChunk=audioBuffer->getMemChunk();
  memChunk->addLock();

  StatusInfo* statusInfo=config->getStatusInfo();
  if (statusInfo->getChange()) {
    statusInfo->copyTo(status);
  }

  freePos++;
  if (overflowDanger() == true) {
    setOverflowDangerLock();
  }
}


DeviceConfig* DeviceConfigFifo::dequeue() {
  lDequeued=true;
  return configQueue[0];
}


void DeviceConfigFifo::forwardQueue() {
  int i;
  DeviceConfig* config;
  MemChunk* memChunk;
  AudioBuffer* audioBuffer;

  // first we define the status of the fifo. by default a fifo
  // is active but if we forward the queue and the deviceConfig
  // say that the status is stopped the fifo gets inactive
  // it stays inactive until a deviceConfig is inserted with
  // playing status

  if (freePos == 0) {
    cout << "forward queue on empty queue called!"<<endl;
    exit(0);
  }
  if (lDequeued == false) {
    cout << "error forward without dequeue"<<endl;
    exit(0);
  }

  lDequeued=false;
  config=configQueue[0];
  audioBuffer=config->getAudioBuffer();
  memChunk=audioBuffer->getMemChunk();
  memChunk->removeLock();

  for(i=0;i<=freePos-1;i++) {
    configQueue[i]=configQueue[i+1];
  }
  configQueue[freePos]=config;
  
  freePos--;
  if (overflowDanger() == false) {
    removeOverflowDangerLock();
  }
}


int DeviceConfigFifo::overflowDanger() {
  return (freePos == (_MAX_FIFO_ENTRIES-2));
}


int DeviceConfigFifo::getFillgrade() {
  return freePos;
}


int DeviceConfigFifo::getWaitForStoppedSource() {
  return lWaitForStoppedSource;
}


void DeviceConfigFifo::setWaitForStoppedSource(int lWaitForStopped) {
  this->lWaitForStoppedSource=lWaitForStopped;
}


int DeviceConfigFifo::getWaitForPausedSource() {
  return lWaitForPausedSource;
}


void DeviceConfigFifo::setWaitForPausedSource(int lWaitForPaused) {
  this->lWaitForPausedSource=lWaitForPaused;
}


int DeviceConfigFifo::getDequeue() {
  return lDequeued;
}


int DeviceConfigFifo::getStatus() {
  return status->getStatus(); 
}


NodeDevice* DeviceConfigFifo::getID() {
  return id;
}


void DeviceConfigFifo::setOverflowDangerLock() {
  if (freePos > 0) {
    DeviceConfig* config=configQueue[0];
    StreamProducer* producer=(StreamProducer*)config->getStreamProducer();
    DaisyChain* daisyChain=(DaisyChain*)(producer->getDaisyChain());
    daisyChain->addLock(this,producer);
  }
}


void DeviceConfigFifo::removeOverflowDangerLock() {
  if (freePos > 0) {
    DeviceConfig* config=configQueue[0];
    StreamProducer* producer=(StreamProducer*)config->getStreamProducer();
    DaisyChain* daisyChain=(DaisyChain*)(producer->getDaisyChain());
    daisyChain->removeLock(this,producer);
  } 
}

  

void DeviceConfigFifo::clear() {
  while(getFillgrade() > 0) {
    dequeue();
    forwardQueue();
  }
  return;
}
