/* qmbrowser.cpp
 *
 * Marius Sundbakken <marius@fix.no>
 *
 * Last modified: Time-stamp: <08-Feb-1999 06:52:04 amos>
 */

#include <iostream.h>

#include <qapplication.h>
#include <qdir.h>
#include <qheader.h>
#include <qimage.h>
#include <qlayout.h>
#include <qsplitter.h>
#include <qtooltip.h>
#include <qwhatsthis.h>

#include "qmbrowser.h"
#include "qmdirtree.h"
#include "qmfilelist.h"
#include "qmlistview.h"
#include "qmcachedpreview.h"
#include "qmtabbar.h"
#include "qminfoitem.h"
#include "strings.h"

/*!
  \class QmBrowser qmbrowser.h
  \brief The QmBrowser class contains dir tree, file list, thumbnail list
  \brief and preview.

  This widget shows a directory tree, a hotlist, a preview and a file list.
*/


/*!
  Creates a new browser.

  The \a Parent and \a Name arguments are sent to the QWidget constructor.
*/

QmBrowser::QmBrowser(
    QWidget *Parent,
    const char *Name)
    : QWidget(Parent, Name)
{
    QBoxLayout *pMainLayout = new QHBoxLayout(this, 2);

    QSplitter *pPreview_ListSplit = new QSplitter(QSplitter::Horizontal, this);
    QSplitter *pFile_PreviewSplit = new QSplitter(QSplitter::Vertical,
						  pPreview_ListSplit);

    pTabBar = new QmTabBar(QTabBar::RoundedAbove, pFile_PreviewSplit);
    {
	pDirList = new QmListView(pTabBar);
	{
	    // We must have at least one column. Make it, then hide the header.
	    pDirList->addColumn("Dir");
	    pDirList->header()->hide();
	
	    pDirList->setFrameStyle(QFrame::WinPanel | QFrame::Sunken);
	    pDirList->setMinimumSize(50, 50);
	    pDirList->setMultiSelection(false);

	    connect(pDirList, SIGNAL(doubleClicked(QListViewItem *)),
		    this, SLOT(dirInTreeSelected(QListViewItem *)));

	    connect(pDirList, SIGNAL(returnPressed(QListViewItem *)),
		    this, SLOT(dirInTreeSelected(QListViewItem *)));

	    connect(pDirList, SIGNAL(selectionChanged(QListViewItem *)),
		    this, SLOT(dirInTreeHighlighted(QListViewItem *)));

	    // The stuff below is pretty slow! Maybe do this after
	    // the window has opened?
	    
	    QmDirectory *root = new QmDirectory(pDirList);
	    root->setOpen(true); // slow
//  	    root->setOpen(false); // fast!
	}

	DirTreeID = pTabBar->addWidget(pDirList, "&Dirs");
	
	pHotList = new QmHotList(pTabBar);
	{
	    pHotList->addColumn("Name");
	    pHotList->addColumn("Path");
	    pHotList->setFrameStyle(QFrame::WinPanel | QFrame::Sunken);
	    
	    QToolTip::add(pHotList, TT_HOT_LIST);
	    QWhatsThis::add(pHotList, WT_HOT_LIST);
	    
	    connect(pHotList, SIGNAL(doubleClicked(QListViewItem *)),
		    this, SLOT(hotlistSelected(QListViewItem *)));

	    connect(pHotList, SIGNAL(returnPressed(QListViewItem *)),
		    this, SLOT(hotlistSelected(QListViewItem *)));
	}

	HotListID = pTabBar->addWidget(pHotList, "H&otlist");
	
	pTabBar->setMinimumSize(100, 100);
    }
    
    pPreviewImage = new QmCachedPreview(pFile_PreviewSplit);
    {
	pPreviewImage->setMinimumSize(50, 50);
	pPreviewImage->setFocusPolicy(QWidget::NoFocus);

	connect(pPreviewImage, SIGNAL(clicked(const QPixmap &)),
		this, SLOT(showImage(const QPixmap &)));

	QWhatsThis::add(pPreviewImage, WT_PREVIEW);
    }


    pFileList = new QmFileList(pPreview_ListSplit);
    {
 	connect(pFileList, SIGNAL(fileHighlighted(const char *)),
 		this, SLOT(showPreview(const char *)));

 	connect(pFileList, SIGNAL(fileSelected(const char *)),
 		this, SLOT(selected(const char *)));

	connect(pFileList, SIGNAL(dirSelected(const char *)),
		this, SLOT(dirInListSelected(const char *)));
    }

    // Add everything. pPreview_ListSplit is nested with other widgets.
    
    pMainLayout->addWidget(pPreview_ListSplit);
    
    pFileList->setPath( QDir::currentDirPath() );

    pDirTimer = new QTimer(this);
    connect(pDirTimer, SIGNAL(timeout()), this, SLOT(dirTimer()));

    pPreviewTimer = new QTimer(this);
    connect(pPreviewTimer, SIGNAL(timeout()), this, SLOT(previewTimer()));

    pMainLayout->activate();
}


/*!

*/

QmBrowser::~QmBrowser()
{
}


/*!

 */

// void
// QmBrowser::focusInEvent(
//     QFocusEvent *e)
// {
//     cout << "QmBrowser::focusInEvent()\n";

//     if(pLastActive == 0)
// 	return;

//     pLastActive->setFocus();

//     QWidget::focusInEvent(e);
// }


/*!

 */

// void
// QmBrowser::focusOutEvent(
//     QFocusEvent *e)
// {
//     cout << "QmBrowser::focusOutEvent()\n";
    
//     pLastActive = focusWidget();

//     QWidget::focusOutEvent(e);
// }


/*!
  Emits the fileSelected() signal with the next image filename.
  Useful for browsing in full image/screen mode.

  \sa queryPrevImage().
*/

void
QmBrowser::queryNextImage()
{
    pFileList->queryNextImage();
}


/*!
  Emits the fileSelected() signal with the previous image filename.
  Useful for browsing in full image/screen mode.

  \sa queryNextImage().
*/

void
QmBrowser::queryPrevImage()
{
    pFileList->queryPrevImage();
}


/*!
  Enables the preview if \a Enable is true, disables and hides the
  preview otherwise.
*/

void
QmBrowser::setPreview(
    bool Enable)
{
    if( ! Enable)
	pPreviewImage->hide();
    else
    {
	pPreviewImage->show();

	// Reload preview in case a different file has been selected
	// since it was last hidden

	/*
	QmInfoItem *pItem = dynamic_cast<QmInfoItem *>(pFileList->currentItem());

 	if(pItem)
 	    showPreview(pItem->fileInfo->filePath());
	*/
    }
}


/*!
  Shows a scaled down version of the image \a Filename.

  This function will do nothing if the preview widget is not visible.

  The preview will not be shown immediately. A timer will be started,
  and after a reasonable delay, the image will be loaded, scaled and shown.
  This way, one can quickly browse around in the file list without having
  to wait before a preview is shown for every highlighted image.
*/

void
QmBrowser::showPreview(
    const char *Filename)
{
    CHECK_PTR(pPreviewTimer);
    CHECK_PTR(pPreviewImage);
    CHECK_PTR(Filename);

    if( ! pPreviewImage->isVisibleToTLW())
	return;
    
    CurrentFileSelection = Filename;
    
    if(pPreviewTimer->isActive())
	pPreviewTimer->stop();
    
    pPreviewTimer->start(ShowPreviewDelay, true);
}


/*!
  Shows the preview image for a previously highlighted image. This
  function is used together with a timer.
*/

void
QmBrowser::previewTimer()
{
    CHECK_PTR(pPreviewTimer);
    CHECK_PTR(pPreviewImage);
    
    pPreviewTimer->stop();

//    emit message("Loading image...");
    
    if( ! ThePixmap.load(CurrentFileSelection))
    {
//	emit message("Failed to load image");
	return;
    }

//    emit message("Done loading image");

    pPreviewImage->setPixmap(ThePixmap);
}


/*!
  Shows the image \a Pixmap in its full size.
*/

void
QmBrowser::showImage(
    const QPixmap &Pixmap)
{
//    cout << "QmBrowser::showImage(): Preview clicked\n";

    // This is ineffective. The preview widget has already loaded the
    // pixmap. A better preview widget (= better image loading) may
    // change this, though.
    
//    emit fileSelected(CurrentFileSelection);
    emit imageSelected(Pixmap);
}


/*!
  Called whenever a file is selected in the file list.
*/

void
QmBrowser::selected(
    const char *Filename)
{
    emit fileSelected(Filename);
}


/*!

 */

void
QmBrowser::hotlistSelected(
    QListViewItem *pItem)
{
    CHECK_PTR(pItem);
    
    pFileList->setPath(pItem->text(QmHotList::PathColumn));
}


/*!
  Lists the directory given by \a pItem.
*/

/*
  Called whenever an item in the directory tree is selected.
*/

void
QmBrowser::dirInTreeSelected(
    QListViewItem *pItem)
{
    CHECK_PTR(pItem);

    // Stop the timer, and list the directory.
    
    dirTimer();

    pDirList->setOpen(pItem, true);
    pFileList->visibleWidget()->setFocus();
}


/*!

*/

/*
  This needs some more work. Primarily, it doesn't react to path changes
  when done in the string widget.

  The dir is listed twice from the file system; once in the listview,
  once in the dirtree. When a dir is selected in the listview, it get's
  read there, then a signal is sent to the treeview, which reloads that
  directory and selects the directory just listed. That will result in
  a dirlisting signal being sent back to the listview (where the dir
  is already listed).
*/

void
QmBrowser::dirInListSelected(
    const char *Dirname)
{
    /*
    cout << "QmBrowser::dirInListSelected(): " << endl
	 << "Selected in list: " << Dirname << endl
	 << "Selected in tree: " << (pDirList->currentItem() == 0 ? "None" : pDirList->currentItem()->text(0))
	 << endl;
    */

    QDir FullDir(Dirname);
    QDir d = FullDir;

    QStrList l;
    l.append(d.dirName());
    
    for(int i = 0; d != QDir::root(); i++)
    {
	d.cdUp();
	l.insert(0, d.dirName());
    }

    /*
    for(int i = 1; i < l.count(); i++)
    {
	cout << "List: " << l.at(i) << endl;
    }
    */

    int j = 1;
    
    for(QListViewItem *i = pDirList->firstChild(); i != 0 && (uint)j < l.count();
	i = i->itemBelow())
    {
//	cout << "Tree: " << i->text(0) << endl;
	
	if(strcmp(l.at(j), i->text(0)) == 0)
	{
//	    cout << "List->Tree: " << l.at(j++) << endl;
	    j++;
	    i->setOpen(true);
//	    i->setSelected(true);
	    pDirList->setSelected(i, true);
	    pDirList->ensureItemVisible(i);
	}
    }
}


/*!

*/

void
QmBrowser::dirInTreeHighlighted(
    QListViewItem *pItem)
{
    QmDirectory *i = dynamic_cast<QmDirectory *>(pItem);

    if(i == 0)
	return;
    
    CurrentDirSelection = i->fullName();

    // To keep a snappy user interface, we have a little delay before
    // listing a directory *highlighted* in the directory tree.
    
    if(pDirTimer->isActive())
	pDirTimer->stop();
    
    pDirTimer->start(ListDirDelay, true);
}


/*!
  Lists the directory set earlier. This function is used together with
  a timer.
*/

void
QmBrowser::dirTimer()
{
    CHECK_PTR(pDirTimer);
    CHECK_PTR(pFileList);
    
    pDirTimer->stop();
    pFileList->setPath(CurrentDirSelection);
}


void QmBrowser::setPath( QString dir )
{
    pFileList->setPath( dir );
}
