/* qmimageview.cpp
 *
 * Marius Sundbakken <marius@fix.no>
 *
 * Last modified: Time-stamp: <08-Feb-1999 06:16:41 amos>
 */

#include <iostream.h>
#include <math.h>

#include <qapplication.h>
#include <qcursor.h>
#include <qdatetime.h>
#include <qdragobject.h>
#include <qfileinfo.h>
#include <qkeycode.h>
#include <qlabel.h>
#include <qmessagebox.h>
#include <qpainter.h>
#include <qtimer.h>

#include "qmimageview.h"
#include "qmmainwindow.h"

/* XPM */
const char * picture_xpm[] = {
"16 16 3 1",
" 	c None",
".	c #000000",
"X	c #0000FF",
"XXXXXXXX        ",
"XXXXXXXX        ",
"XXXXXXXX        ",
"XXXXXXXX   X    ",
"XXXXXXXX  XXX   ",
"         XXXXX  ",
"        XXXXXXX ",
"       XXXXXXXXX",
"   XXX          ",
" XXXXXXX        ",
"XXXXXXXXX       ",
"XXXXXXXXX       ",
"XXXXXXXXX       ",
" XXXXXXX        ",
"   XXX          ",
"                "};

extern QLabel *g_pLargeInfo;
extern QLabel *g_pSmallInfo;

/*
 TODO:

 Hide cursor if it has been still for more than, say, 3 seconds.
  
 */

/*!
  \class QmImageView qmimageview.h
  \brief The QmImageView class shows a picture in a scrollable area.

  This class is meant for showing images in their original size.

  \bug The keyboard scrolling is buggy.
  \bug The drag and drop of "itself" must be fixed (i.e. not accept drops from
       itself.)
*/

/*!

*/

QmImageView::QmImageView(
    QWidget *Parent,
    const char *Name)
    : QScrollView(Parent, Name),
      QDropSite(this->viewport())
{
    ButtonReleased = false;

    setVScrollBarMode(AlwaysOff);
    setHScrollBarMode(AlwaysOff);

    BlackBackground.setStyle(SolidPattern);
    BlackBackground.setColor(black);
//  	viewport()->setBackgroundMode( NoBackground );

    setFocusPolicy(StrongFocus);

    for(int i = 0; i < 4; i++)
    {
	DirectionPressed[i] = false;
    }
    
    ShiftPressed = false;

    pTimer = new QTimer(this);
    connect(pTimer, SIGNAL(timeout()), this, SLOT(timerDone()));

    // Check QWidget::setBackgroundMode(NoBackground) if flickering
    // becomes a problem.

//    setBackgroundMode(NoBackground);
//    setBackgroundColor(black);
}


/*!

*/

QmImageView::~QmImageView()
{
}


/*!
  Loads the image given with \a Filename. Returns true if successful,
  false otherwise.  
*/

bool
QmImageView::loadImage(
    const char *Filename)
{
    CurrentZoom = 100;
    ImageFilename = Filename;

    QTime t;
    t.start();
    
    if( ! ThePixmap.load(Filename))
	return false;

    PixmapToShow = ThePixmap;
    
    setContentsPos(0, 0);
    resizeContents(PixmapToShow.width(), PixmapToShow.height());

    // Repaint the _drawing_ area
    
    viewport()->repaint();

    CHECK_PTR(g_pLargeInfo);
    CHECK_PTR(g_pSmallInfo);

    QString s;
    s.sprintf("%dx%d", ThePixmap.width(), ThePixmap.height());
    g_pSmallInfo->setText((const char *)s);

    QFileInfo fi(Filename);
    s.sprintf("%s (%s format) loaded in %0.1f second(s)",
	      (const char *)fi.fileName(),
//	      (ThePixmap.depth() == 32 ? 24 : ThePixmap.depth()),
	      QImage::imageFormat(Filename),
	      (float)t.elapsed() / 1000.0);
    g_pLargeInfo->setText((const char *)s);

    return true;
}


void
QmImageView::setPixmap(
    const QPixmap &Pixmap)
{
    CurrentZoom = 100;
    ThePixmap = Pixmap;
    PixmapToShow = ThePixmap;
    setContentsPos(0, 0);
    resizeContents(PixmapToShow.width(), PixmapToShow.height());

    // Repaint the _drawing_ area
    
    viewport()->repaint();
}


int
QmImageView::currentZoom(void) const
{
    return CurrentZoom;
}


/*!
  Zooms the image by \a Percent percent. If \a Percent is negative, the
  image is scaled down by that percent.
*/

void
QmImageView::zoom(
    int Percent)
{
    ASSERT(Percent <= 300);
    ASSERT(Percent >= 10);

    CurrentZoom = Percent;
    
    double Zoom = Percent / 100.0;

    QWMatrix m;
    m.scale(Zoom, Zoom);
    PixmapToShow = ThePixmap.xForm(m);
   
    resizeContents(PixmapToShow.width(), PixmapToShow.height());

    // Repaint the _drawing_ area
    viewport()->repaint();
}


/*!

 */

void
QmImageView::keyPressEvent(
    QKeyEvent *e)
{
    switch(e->key())
    {
    case Key_Down:
 	VScrollDir = VDown;
	DirectionPressed[0] = true;
	
 	if( ! pTimer->isActive())
 	    pTimer->start(ScrollDelay, true);
	break;
	
    case Key_Up:
	VScrollDir = VUp;
	DirectionPressed[1] = true;
	
	if( ! pTimer->isActive())
	    pTimer->start(ScrollDelay, true);
	break;
	
    case Key_Left:
	HScrollDir = HLeft;
	DirectionPressed[2] = true;
	
	if( ! pTimer->isActive())
	    pTimer->start(ScrollDelay, true);
	break;

    case Key_Right:
	HScrollDir = HRight;
	DirectionPressed[3] = true;
	
	if( ! pTimer->isActive())
	    pTimer->start(ScrollDelay, true);
	break;

    case Key_Shift:
	ShiftPressed = true;
	break;
	    
    case Key_Delete:
	deleteImage();
	break;

    case Key_Space:
    case Key_PageDown:
	emit queryNextImage();
	break;

    case Key_Backspace:
    case Key_PageUp:
	emit queryPrevImage();
	break;

    case Key_Enter:
    case Key_Return:
	emit returnPressed();
	break;
	
    default:
	e->ignore();
    }
}


/*!

*/

void
QmImageView::keyReleaseEvent(
    QKeyEvent *e)
{
    switch(e->key())
    {
    case Key_Down:
	DirectionPressed[0] = false;
	
	if( ! directionPressed())
	{
	    if(pTimer->isActive())
		pTimer->stop();
	    
	    VScrollDir = VNone;
	}
	else if(DirectionPressed[1] == true)
	    VScrollDir = VUp;
	else
	    VScrollDir = VNone;
	
	break;
	
    case Key_Up:
	DirectionPressed[1] = false;
	
	if( ! directionPressed())
	{
	    if(pTimer->isActive())
		pTimer->stop();
	    
	    VScrollDir = VNone;
	}
	else if(DirectionPressed[0] == true)
	    VScrollDir = VDown;
	else
	    VScrollDir = VNone;
	
	break;
	
    case Key_Left:
	DirectionPressed[2] = false;
	
	if( ! directionPressed())
	{
	    if(pTimer->isActive())
		pTimer->stop();
	    
	    HScrollDir = HNone;
	}
	else if(DirectionPressed[3] == true)
	    HScrollDir = HRight;
	else
	    HScrollDir = HNone;
	
	break;

    case Key_Right:
	DirectionPressed[3] = false;
	
	if( ! directionPressed())
	{
	    if(pTimer->isActive())
		pTimer->stop();
	    
	    HScrollDir = HNone;
	}
	else if(DirectionPressed[2] == true)
	    HScrollDir = HLeft;
	else
	    HScrollDir = HNone;
	
	break;

    case Key_Shift:
	ShiftPressed = false;
	
	break;

    default:
	e->ignore();
    }
}


/*!
  Returns true if a scroll key is pressed, false if not.
*/

bool
QmImageView::directionPressed(void) const
{
    for(int i = 0; i < 4; i++)
    {
	if(DirectionPressed[i] == true)
	    return true;
    }
    
    return false;
}

/*!

*/

void
QmImageView::leaveEvent(
    QEvent * /*e*/)
{
//     if( ! ButtonReleased)
//     {
//  	QImageDrag *d = new QImageDrag(TheImage, viewport());
//  	d->setPixmap(QPixmap(picture_xpm), QPoint(8, 8));
//  	d->dragCopy();
//     }
}


/*!

*/

void
QmImageView::viewportMousePressEvent(
    QMouseEvent *e)
{
    if(e->button() == LeftButton)
    {
	// WARNING: QImageDrag does not take pixmaps as argument.
	// Converting from pixmap to image is slow. Write a QmPixmapDrag?

	
  	QImageDrag *d = new QImageDrag(ThePixmap.convertToImage(), viewport());
  	d->setPixmap(QPixmap(picture_xpm), QPoint(8, 8));
  	d->dragCopy();
	
	if(contentsWidth() > width() && contentsHeight() > height())
	    setCursor(sizeAllCursor);
	else if(contentsWidth() <= width() && contentsHeight() > height())
	    setCursor(sizeVerCursor);
	else if(contentsWidth() > width() && contentsHeight() <= height())
	    setCursor(sizeHorCursor);
    }
}


/*!

*/

void
QmImageView::viewportMouseReleaseEvent(
    QMouseEvent *)
{
    setCursor(arrowCursor);

    ButtonReleased = true;
}


/*!

*/

void
QmImageView::viewportMouseMoveEvent(
    QMouseEvent *e)
{
    static preX = e->x();
    static preY = e->y();

    int dx, dy;

    // We need to check if this is the first move after a new mouse click
    // has been done. This is so that delta x and delta y is to be correct.
    // If we don't check this, delta x and delta y will jump to great
    // numbers whenever the mouse is moved and pressed elsewhere.
    
    if(ButtonReleased)
    {
	dx = 0;
	dy = 0;

	ButtonReleased = false;
    }
    else
    {
	dx = e->x() - preX;
	dy = e->y() - preY;
    }

    // Multiply by 2 to speed up the scrolling
    
    scrollBy( -dx * 2, -dy * 2);

    preX = e->x();
    preY = e->y();
}


/*!
  Draws the image.
*/

void
QmImageView::drawContentsOffset(
    QPainter *p,
    int OffsetX, int OffsetY,
    int ClipX, int ClipY,
    int ClipW, int ClipH)
{
    p->setClipRect(ClipX - OffsetX, ClipY - OffsetY, ClipW, ClipH);

    // Fill the background with black. This is probably not the fastest
    // way, and it may cause some flickering.
    
//      p->fillRect(ClipX-OffsetX, ClipY-OffsetY, ClipW, ClipH, BlackBackground);

    // Draw the pixmap. Center it horizontally and vertically if th
    // image is smaller than the widget.

    int XPos;
    int YPos;

    if(PixmapToShow.width() > width())
		XPos = - OffsetX;
    else
		XPos = width() / 2 - PixmapToShow.width() / 2 - OffsetX;

    if(PixmapToShow.height() > height())
		YPos = - OffsetY;
    else
		YPos = height() / 2 - PixmapToShow.height() / 2 - OffsetY;
    
		// Needs to be fixed
//  	QRect dr( QPoint( XPos, YPos ), PixmapToShow.size() );
//  	QRect cr( QPoint( ClipX - OffsetX, ClipY - OffsetY), QSize( ClipW, ClipH ) );
//  	QRect nr = cr.intersect( dr );

//  	if ( !nr.isEmpty() )
//  		p->drawPixmap( cr.topLeft(), PixmapToShow );

	p->drawPixmap( XPos, YPos, PixmapToShow );
}


/*!

 */

void
QmImageView::dragEnterEvent(
    QDragEnterEvent *e)
{
    cout << "QmImageView::dragEnterEvent()\n";
    
    // Check if you want the drag...
    
    if(QTextDrag::canDecode( e ) || QImageDrag::canDecode( e ) )
    {
	e->accept();
    }


    // Give the user some feedback...
    
    QString t;
    for( int i=0; e->format( i ); i++ ) {
	if ( *(e->format( i )) ) {
	    if ( !t.isEmpty() )
		t += "\n";
	    t += e->format( i );
	}
    }
    
//     emit message( t );
//     setBackgroundColor(white);
}


/*!

 */

void
QmImageView::dragMoveEvent(
    QDragMoveEvent */*e*/)
{
//    cout << "QmImageView::dragMoveEvent()\n";
}


/*!

 */

void
QmImageView::dragLeaveEvent(
    QDragLeaveEvent * /*e*/)
{
//  cout << "QmImageView::dragLeaveEvent()\n";
    
    // Give the user some feedback...
//     emit message("");
//     setBackgroundColor(lightGray);
}


/*!

 */

void
QmImageView::dropEvent(
    QDropEvent *e)
{
    cout << "QmImageView::dropEvent()\n";

    // Try to decode to the data you understand...

    QString str;
    if(QTextDrag::decode(e, str))
    {
	// Does the file exist (is it a file)?
	
	if(QFile::exists(str))
	{
	    QImage img;

	    if(img.load(str))
	    {
		ThePixmap = img;

		resizeContents(PixmapToShow.width(), PixmapToShow.height());

		// Repaint the _drawing_ area
    
		viewport()->repaint();
	    }
	}
	
	return;
    }

    QImage img;
    if(QImageDrag::decode(e, img))
    {
	ThePixmap = img;
	resizeContents(PixmapToShow.width(), PixmapToShow.height());
	viewport()->repaint();
	return;
    }
}


/*!
  This function is necessary for smooth scrolling when using the
  keyboard. This is because we don't get the key event often enough
  for a smooth scroll. So, instead, we scroll some pixels between every
  time we receive an event (and not all in one chunk).

  The scrolling isn't very smooth, but it's better than just calling
  scrollBy() in keyPressEvent();
*/

void
QmImageView::timerDone()
{
    int DeltaX, DeltaY, PixVal;

    if(ShiftPressed)
	PixVal = PixelsToScroll;
    else
	PixVal = PixelsToScroll * 3;

    switch(VScrollDir)
    {
    case VDown:
	DeltaY = PixVal;
	break;
	
    case VUp:
	DeltaY = -PixVal;
	break;
	
    case VNone:
    default:
	DeltaY = 0;
	break;
    }
		
    switch(HScrollDir)
    {
    case HLeft:
	DeltaX = -PixVal;
	break;
	
    case HRight:
	DeltaX = PixVal;
	break;
	
    case HNone:
    default:
	DeltaX = 0;
	break;
    }
    
    scrollBy(DeltaX, DeltaY);

    pTimer->start(ScrollDelay, true);
//    pTimer->stop();
}

void
QmImageView::deleteImage(void)
{
    QmMainWindow::deleteItem((const char *)ImageFilename);
}

