/*  x11amp - graphically mp3 player..
 *  Copyright (C) 1998-1999  Mikael Alm, Olle Hallnas, Thomas Nilsson and 4Front Technologies
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/
#include "x11amp.h"

pthread_mutex_t vis_mutex=PTHREAD_MUTEX_INITIALIZER;

struct InputPluginData
{
	GList 		*input_list;
	InputPlugin	*current_input_plugin;
	gboolean	playing;
	gboolean	paused;
};


typedef struct
{
	gint	time;
	guchar	spectrum[75];
} SpectrumEntry;
		
		
struct InputPluginData *ip_data;
GList *spectrum_list=NULL;
gchar *input_info_text=NULL;
extern PlayStatus *mainwin_playstatus;


void input_set_spectrum(int time,unsigned char *s);
void input_set_info(char *t,int len,int r,int f,int nch);
void input_set_info_text(gchar *text);	

InputPlugin *get_current_input_plugin(void)
{
	return ip_data->current_input_plugin;
}

void set_current_input_plugin(InputPlugin *ip)
{
	ip_data->current_input_plugin=ip;
}

void add_input_plugin(gchar *filename)
{
	void *h;
	InputPlugin *(*gpi)(void);
	InputPlugin *p;
	
	if(h=dlopen(filename,RTLD_NOW))
	{
		gpi=dlsym(h,"get_iplugin_info");
		if(gpi)
		{
			p=gpi();
			p->handle=h;
			p->filename=filename;
			p->set_spectrum=input_set_spectrum;
			p->set_info=playlist_set_info;
			p->set_info_text=input_set_info_text;

			if(p->init) 
				p->init();
			
			ip_data->input_list=g_list_prepend(ip_data->input_list,p);
		}
	}
	else
		fprintf(stderr,"%s\n",dlerror());
}

void scan_input_plugins(char *dirname)
{
	gchar *filename,*ext;
	DIR *dir;
	struct dirent *ent;
	struct stat statbuf;
	
	dir=opendir(dirname);
	if(dir)
	{
		while(ent=readdir(dir))
		{
			filename=(gchar *)g_malloc(strlen(dirname)+strlen(ent->d_name)+2);
			sprintf(filename,"%s/%s",dirname,ent->d_name);
			if(!stat(filename,&statbuf))
			{
				if(S_ISREG(statbuf.st_mode))
				{
					if(ext=strrchr(ent->d_name,'.'))
					{
						if(!strcmp(ext,".so"))
							add_input_plugin(filename);
						else
							g_free(filename);
					}
					else
						g_free(filename);
				}
				else
					g_free(filename);
			}
			else
				g_free(filename);
		}
	}
}
	
GList *get_input_list(void)
{
	return ip_data->input_list;
}

void input_set_spectrum(int time,unsigned char *s)
{
	SpectrumEntry *spectrum;
	
	pthread_mutex_lock(&vis_mutex);
	spectrum=g_malloc0(sizeof(SpectrumEntry));
	spectrum->time=time;
	memcpy(spectrum->spectrum,s,75);
	spectrum_list=g_list_append(spectrum_list,spectrum);
	pthread_mutex_unlock(&vis_mutex);
}
	
static int inputlist_compare_func(const void *a, const void *b)
{
	return strcasecmp ((*(InputPlugin **)a)->description, (*(InputPlugin **)b)->description);
}

void init_input_plugins(void)
{
	gchar *dir;

	ip_data=g_malloc0(sizeof(struct InputPluginData));
	dir=g_strconcat(X11AMP_DIR,"/Input",NULL);
	scan_input_plugins(dir);
	g_free(dir);
	ip_data->input_list=sort_glist(ip_data->input_list,inputlist_compare_func);
	
}

void cleanup_input_plugins(void)
{
	InputPlugin *ip;
	GList *node;
	
	
	if(get_input_playing())
		input_stop();
	
	node=get_input_list();
	while(node)
	{
		ip=(InputPlugin *)node->data;
		dlclose(ip->handle);
		node=node->next;
	}
	
}
		

void input_play(char *filename)
{
	GList *node;
	InputPlugin *ip;

	
	node=get_input_list();
	while(node)
	{

		ip=(InputPlugin *)node->data;
		if(ip->is_our_file(filename))
		{
			set_current_input_plugin(ip);
			ip->output=get_current_output_plugin();
			ip->play_file(filename);
			ip_data->playing=TRUE;
			break;

		}
		node=node->next;
	}
}

void input_seek(int time)
{
	GList *node;
	if(ip_data->playing)
	{
		pthread_mutex_lock(&vis_mutex);
		node=spectrum_list;
		while(node)
		{
			g_free(node->data);
			node=g_list_next(node);
		}
		g_list_free(spectrum_list);
		spectrum_list=NULL;
		pthread_mutex_unlock(&vis_mutex);
		get_current_input_plugin()->seek(time);
	}
}

void input_stop(void)
{
	GList *node;
	fflush(stdout);
	if(ip_data->playing)
	{
		if(ip_data->paused)
			input_pause();
		get_current_input_plugin()->stop();
		pthread_mutex_lock(&vis_mutex);
		node=spectrum_list;
		while(node)
		{
			g_free(node->data);
			node=g_list_next(node);
		}
		g_list_free(spectrum_list);
		spectrum_list=NULL;
		pthread_mutex_unlock(&vis_mutex);
		ip_data->playing=FALSE;
		ip_data->paused=FALSE;
		if(input_info_text)
		{
			g_free(input_info_text);
			input_info_text=NULL;
			mainwin_set_info_text();
		}
	}
}

void input_pause(void)
{
	if(get_input_playing())
	{
		ip_data->paused=!ip_data->paused;
		if(ip_data->paused)
			playstatus_set_status(mainwin_playstatus,STATUS_PAUSE);
		else
			playstatus_set_status(mainwin_playstatus,STATUS_PLAY);
		get_current_input_plugin()->pause(ip_data->paused);	
	}
}

gint input_get_time(void)
{
	gint time;
	if(get_input_playing())
		time=get_current_input_plugin()->get_time();
	else
		time=-1;
	return time;
}

void input_set_eq(int on,float preamp,float *bands)
{
	if(ip_data->playing)
		if(get_current_input_plugin()->set_eq)
			get_current_input_plugin()->set_eq(on,preamp,bands);
}

gboolean get_input_playing(void)
{
	return ip_data->playing;
}

guchar *input_get_spectrum(gint time)
{
	GList *node,*prev;
	SpectrumEntry *spectrum;
	guchar *ret=NULL;
	gboolean found;
	
	pthread_mutex_lock(&vis_mutex);
	node=spectrum_list;
	found=FALSE;
	while(g_list_next(node)&&!found)
	{
		spectrum=(SpectrumEntry *)node->data;
		if(((SpectrumEntry *)g_list_next(node)->data)->time>=time&&spectrum->time<time)
		{
			ret=g_malloc(75);
			memcpy(ret,spectrum->spectrum,75);
			found=TRUE;
			while(node)
			{
				g_free(node->data);
				prev=g_list_previous(node);
				spectrum_list=g_list_remove_link(spectrum_list,node);
				g_list_free_1(node);
				node=prev;
			}
		}
		else
			
			node=g_list_next(node);
	}
	pthread_mutex_unlock(&vis_mutex);
	return ret;
}

gchar *input_get_info_text(void)
{
	return input_info_text;
}

void input_set_info_text(gchar *text)
{
	if(input_info_text)
		g_free(input_info_text);
	if(text)
		input_info_text=g_strdup(text);
	else
		input_info_text=NULL;
	mainwin_set_info_text();
}

void input_about(gint index)
{
	InputPlugin *ip;
	
	ip=g_list_nth(ip_data->input_list,index)->data;
	if(ip->about)
		ip->about();
}

void input_configure(gint index)
{
	InputPlugin *ip;
	
	ip=g_list_nth(ip_data->input_list,index)->data;
	if(ip->configure)
		ip->configure();
}
